import typing
from dico import Message, Embed, AllowedMentions, MessageReference, Component
from dico.model.extras import FILE_TYPE


class Context(Message):
    def __init__(self, client, resp, prefix, command, name_used, **kwargs):
        super().__init__(client, resp, **kwargs)
        self.prefix = prefix
        self.command = command
        self.name_used = name_used
        self.subcommand_name = None

    @classmethod
    def from_message(cls, message: Message, prefix, command, name_used):
        return cls(message.client, message.raw, prefix, command, name_used)

    @property
    def bot(self):
        return self.client

    def send(self,
             content: str = None,
             *,
             embed: typing.Union[Embed, dict] = None,
             embeds: typing.List[typing.Union[Embed, dict]] = None,
             file: FILE_TYPE = None,
             files: typing.List[FILE_TYPE] = None,
             tts: bool = False,
             allowed_mentions: typing.Union[AllowedMentions, dict] = None,
             message_reference: typing.Union[Message, MessageReference, dict] = None,
             component: typing.Union[dict, Component] = None,
             components: typing.List[typing.Union[dict, Component]] = None):
        return self.client.create_message(self.channel,
                                          content,
                                          embed=embed,
                                          embeds=embeds,
                                          file=file,
                                          files=files,
                                          tts=tts,
                                          allowed_mentions=allowed_mentions,
                                          message_reference=message_reference,
                                          component=component,
                                          components=components)



from dico.exception import DicoException


class CommandException(DicoException):
    """Base exception for dico-command."""
    def __init__(self, *args, **fmt):
        super().__init__(*args or [self.__doc__.format(**fmt)])


class CommandAlreadyExists(CommandException):
    """Command {name} already exists."""


class CommandNotExists(CommandException):
    """Command {name} doesn't exist."""


class InvalidArgument(CommandException):
    """Parsing argument has failed."""


class CheckFailed(CommandException):
    """Command check has failed."""


class InvalidModule(CommandException):
    """Unable to find module {path}. Check if path is correct."""


class ModuleAlreadyLoaded(CommandException):
    """Module {path} is already loaded."""


class ModuleNotLoaded(CommandException):
    """Module {path} is never loaded."""


class MissingLoadFunction(CommandException):
    """Module {path} is missing "load" function."""


class MissingUnloadFunction(CommandException):
    """Module {path} is missing "unload" function."""


class AddonAlreadyLoaded(CommandException):
    """Addon {name} is already loaded."""



import typing
import inspect


def is_coro(coro):
    return inspect.iscoroutinefunction(coro) or inspect.isawaitable(coro) or inspect.iscoroutine(coro)


def read_function(func):
    params = [*inspect.signature(func).parameters.values()]
    if params[0].name in ["self", "cls"]:
        del params[0]  # Skip self or cls
    del params[0]  # skip Context
    ret = {}
    for x in params:
        ret[x.name] = {
            "required": x.default == inspect._empty,  # noqa
            "default": x.default,
            "annotation": x.annotation,
            "kind": x.kind
        }
    return ret


def smart_split(ipt: str, args_data: dict, splitter: str = " ", subcommand: bool = False) -> typing.Tuple[list, dict]:
    if len(args_data) == 0:
        if subcommand and ipt:
            return [*ipt.split(splitter)], {}
        return [], {}
    raw_split = ipt.split(splitter)
    # TODO: handle "..."
    initial_split = raw_split
    args_name = [*args_data.keys()]
    last_arg = args_data[args_name[-1]]
    var_positional_in = True in [*map(lambda n: n["kind"] == n["kind"].VAR_POSITIONAL, args_data.values())]
    keyword_only_count = len([x for x in args_data.values() if x["kind"] == x["kind"].KEYWORD_ONLY])
    if len(args_data) == 1:
        if last_arg["kind"] == last_arg["kind"].VAR_POSITIONAL:
            return [ipt], {}
        elif last_arg["kind"] == last_arg["kind"].KEYWORD_ONLY:
            return [], {args_name[-1]: ipt}
        else:
            return [initial_split[0]], {}
    if (len(initial_split) == len(args_data) and not keyword_only_count) or last_arg["kind"] == last_arg["kind"].VAR_POSITIONAL:  # assuming this matches
        return initial_split, {}
    if len(initial_split) != len(args_data) and not var_positional_in and not keyword_only_count:
        raise ValueError("argument count does not match.")
    if keyword_only_count > 1:
        raise AttributeError("maximum keyword-only param number is 1.")
    args = []
    kwargs = {}
    if not ipt.replace(" ", ""):
        raise ValueError("empty input.")
    for i, x in enumerate(args_data.items()):
        k, v = x
        if v["kind"] == v["kind"].KEYWORD_ONLY:
            if var_positional_in:
                raise AttributeError("unable to mix positional-only and keyword-only params.")
            kwargs[k] = ipt or None
            # TODO: fix kwargs is added even if it is not present
            break
        args.append(initial_split[i])
        ipt = ipt.split(initial_split[i], 1)[-1].lstrip()
    return args, kwargs



"""
dico_command
~~~~~~~~~~~~~~~~~~~~~~~~
Command handler for dico.
:copyright: (c) 2021 eunwoo1104
:license: MIT
"""

__version__ = "0.0.6"

from .addon import *
from .bot import *
from .command import *
from .context import *
from .decorator import *



import typing
import asyncio
import inspect
import logging
import traceback
import importlib
import dico
from .command import Command
from .context import Context
from .exception import *
from .utils import smart_split, is_coro

if typing.TYPE_CHECKING:
    from .addon import Addon


class Bot(dico.Client):
    def __init__(self,
                 token: str,
                 prefix: typing.Union[str, typing.List[str], typing.Callable[[dico.Message], typing.Union[typing.Awaitable[str], str, typing.List[typing.Union[typing.Awaitable[str], str]]]]],
                 *,
                 intents: dico.Intents = dico.Intents.no_privileged(),
                 default_allowed_mentions: typing.Optional[dico.AllowedMentions] = None,
                 loop: typing.Optional[asyncio.AbstractEventLoop] = None,
                 cache: bool = True,
                 application_id: typing.Optional[dico.Snowflake.TYPING] = None,
                 monoshard: bool = False,
                 shard_count: typing.Optional[int] = None,
                 **cache_max_sizes: int):
        super().__init__(token,
                         intents=intents,
                         default_allowed_mentions=default_allowed_mentions,
                         loop=loop,
                         cache=cache,
                         application_id=application_id,
                         monoshard=monoshard,
                         shard_count=shard_count,
                         **cache_max_sizes)
        self.prefixes = [prefix] if not isinstance(prefix, list) else prefix
        self.commands = {}
        self.aliases = {}
        self.logger = logging.Logger("dico_command")
        self.on("MESSAGE_CREATE", self.execute_handler)
        self.addons: typing.List[Addon] = []
        self.addon_names: typing.List[str] = []
        self.modules: typing.List[str] = []

    async def get_owners(self) -> typing.List[dico.Snowflake]:
        if not self.application:
            await self.request_current_bot_application_information()
        return self.application.owner_ids if self.application.owner_ids else [self.application.owner.id]

    async def verify_prefix(self, message: dico.Message):
        # final_prefixes = [(await x(message)) if is_coro(x) else x(message) if inspect.isfunction(x) else x for x in self.prefixes]
        final_prefixes = []
        for x in self.prefixes:
            if is_coro(x):
                resp = await x(message)
                if isinstance(resp, str):
                    final_prefixes.append(resp)
                else:
                    final_prefixes.extend(resp)
            elif inspect.isfunction(x):
                resp = x(message)
                if isinstance(resp, str):
                    final_prefixes.append(resp)
                else:
                    final_prefixes.extend(resp)  # noqa
            else:
                final_prefixes.append(x)
        prefix_result = [*map(lambda n: message.content.startswith(n), final_prefixes)]
        if len(set(prefix_result)) != 2 and False in prefix_result:
            return
        for i, r in enumerate(prefix_result):
            if r is True:
                return final_prefixes[i]

    async def execute_handler(self, message: dico.Message):
        if message.author.bot:
            return
        cont = message.content
        if not cont:
            return
        prefix_result = await self.verify_prefix(message)
        if prefix_result is None:
            return
        raw_ipt = cont[len(prefix_result):]
        if not raw_ipt:
            return
        ipt = raw_ipt.split(maxsplit=1)
        if not ipt:
            return
        name = ipt[0]
        cmd = self.commands.get(self.aliases.get(name, name))
        if not cmd:
            return
        context = Context.from_message(message, prefix_result, cmd, name)
        try:
            try:
                args, kwargs = smart_split(ipt[1] if len(ipt) > 1 else "", cmd.args_data, subcommand=bool(cmd.subcommands))
            except Exception as ex:
                raise InvalidArgument from ex
            self.logger.debug(f"Command {name} executed.")
            await cmd.invoke(context, *args, **kwargs)
        except Exception as ex:
            await self.handle_command_error(context, ex)

    def add_command(self, command: Command):
        if command.name in self.commands:
            raise CommandAlreadyExists(name=command.name)
        self.commands[command.name] = command
        for x in command.aliases:
            if x in self.aliases:
                raise CommandAlreadyExists(name=x)
            self.aliases[x] = command.name
        return command

    def remove_command(self, name: str):
        if name not in self.commands:
            return
        command = self.commands.pop(name)
        for x in command.aliases:
            if x in self.aliases:
                del self.aliases[x]

    def command(self, name: typing.Optional[str] = None, *, aliases: typing.Optional[typing.List[str]] = None):
        def wrap(func):
            cmd = Command(func, name or func.__name__, aliases=aliases)
            self.add_command(cmd)
            return cmd
        return wrap

    async def handle_command_error(self, context, ex):
        if await context.command.execute_error_handler(context, ex):
            return
        if context.command.addon and await context.command.addon.on_addon_command_error(context, ex):
            return
        if not self.events.get("COMMAND_ERROR"):
            self.logger.error(f"Error while executing command '{context.command.name}':\n"+''.join(traceback.format_exception(type(ex), ex, ex.__traceback__)))
        else:
            self.dispatch("command_error", context, ex)

    def load_addons(self, *addons: typing.Type["Addon"]):
        for x in addons:
            if x.name in self.addon_names:
                raise AddonAlreadyLoaded(name=x.name)
            self.addon_names.append(x.name)
            loaded = x(self)
            self.addons.append(loaded)
            for c in loaded.commands:
                c.register_addon(loaded)
                self.add_command(c)
            for e in loaded.listeners:
                e.register_addon(loaded)
                self.on_(e.event, e.func)
            if hasattr(self, "interaction"):
                for t in loaded.interactions:
                    t.register_self_or_cls(loaded)
                    self.interaction.add_command(t)
                for cc in loaded.callbacks:
                    cc.register_self_or_cls(loaded)
                    self.interaction.add_callback(cc)

    def unload_addons(self, *addons: typing.Union[str, typing.Type["Addon"]]):
        for x in addons:
            tgt = x if isinstance(x, str) else x.name
            for i, n in enumerate(self.addon_names):
                if n == tgt:
                    del self.addon_names[i]
                    addon = self.addons.pop(i)
                    for c in addon.commands:
                        self.remove_command(c.name)
                    for e in addon.listeners:
                        event_name = e.event.upper().lstrip("ON_")
                        if self.events.get(event_name):
                            self.events.remove(event_name, e.func)
                    if hasattr(self, "interaction"):
                        for t in addon.interactions:
                            self.interaction.remove_command(t)
                        for cc in addon.callbacks:
                            self.interaction.remove_callback(cc)
                    addon.on_unload()

    def load_module(self, import_path: str):
        try:
            module = importlib.import_module(import_path)
            importlib.reload(module)
            if module.__name__ in self.modules:
                raise ModuleAlreadyLoaded(path=import_path)
            self.modules.append(module.__name__)
            if hasattr(module, "load"):
                module.load(self)
            else:
                raise MissingLoadFunction(path=import_path)
        except ImportError:
            raise InvalidModule(path=import_path)

    def unload_module(self, import_path: str):
        try:
            module = importlib.import_module(import_path)
            if module.__name__ in self.modules:
                if hasattr(module, "unload"):
                    module.unload(self)
                    self.modules.remove(module.__name__)
                else:
                    raise MissingUnloadFunction(path=import_path)
            else:
                raise ModuleNotLoaded(path=import_path)
        except ImportError:
            raise InvalidModule(path=import_path)

    def reload_module(self, import_path: str):
        self.unload_module(import_path)
        self.load_module(import_path)



import typing
from .command import Command
from .context import Context


def checks(*funcs: typing.Callable[[Context], typing.Union[bool, typing.Awaitable[bool]]]):
    def wrap(maybe_cmd):
        if isinstance(maybe_cmd, Command):
            maybe_cmd.checks.extend(funcs)
        else:
            if hasattr(maybe_cmd, "_checks"):
                maybe_cmd._checks.extend(funcs)
            else:
                maybe_cmd._checks = [*funcs]
        return maybe_cmd
    return wrap



import typing
from .context import Context
from .exception import CheckFailed, InvalidArgument
from .utils import read_function, is_coro, smart_split


class Command:
    def __init__(self,
                 func,
                 name: str,
                 checks: typing.Optional[typing.List[typing.Callable[[Context], typing.Union[bool, typing.Awaitable[bool]]]]] = None,
                 aliases: typing.Optional[typing.List[str]] = None,
                 is_subcommand: bool = False):
        self.func = func
        self.name = name
        self.checks = checks or []
        self.aliases = aliases or []
        self.subcommands = {}
        self.error_handler = None

        self.args_data = read_function(self.func)
        if hasattr(func, "_checks"):
            self.checks.extend(func._checks)
        self.addon = None
        self.is_subcommand = is_subcommand

    def subcommand(self, *args, **kwargs):
        def wrap(coro):
            cmd = command(*args, **kwargs)(coro)
            cmd.is_subcommand = True
            self.subcommands[cmd.name] = cmd
            return cmd
        return wrap

    def register_addon(self, addon):
        self.addon = addon
        for x in self.subcommands.values():
            x.register_addon(addon)

    async def execute_error_handler(self, ctx, ex):
        if not self.error_handler:
            return False
        args = (self.addon, ctx, ex) if self.addon else (ctx, ex)
        return await self.error_handler(*args)

    def on_error(self, coro):
        self.error_handler = coro
        return coro

    async def evaluate_checks(self, ctx: Context):
        if self.addon and not await self.addon.addon_check(ctx):
            return False
        resp = [n for n in [(await x(ctx)) if is_coro(x) else x(ctx) for x in self.checks] if not n]
        return not resp

    async def invoke(self, ctx: Context, *args, **kwargs):
        if not await self.evaluate_checks(ctx):
            raise CheckFailed
        tgt = self.func
        args = [*args]
        subcommand_invoking = False
        subcommand_name = None
        subcommand = None
        if self.subcommands:
            if args and args[0] in self.subcommands:
                subcommand = self.subcommands[args[0]]
                tgt = subcommand.invoke
                del args[0]
                subcommand_invoking = True
            elif kwargs and [*kwargs.values()][0] in self.subcommands:
                subcommand = kwargs.pop([*kwargs.keys()][0])
                tgt = subcommand.invoke
                subcommand_invoking = True
            elif kwargs or args:
                raise InvalidArgument("unknown subcommand or invalid argument passed.")
        elif (args or kwargs) and not self.args_data:
            raise InvalidArgument("invalid argument data.")
        if subcommand_invoking:
            ctx.subcommand_name = subcommand_name
            msg = ctx.content
            ipt = msg.split(maxsplit=1)
            ipt = ipt[1].split(maxsplit=1) if len(ipt) > 1 else []
            args, kwargs = smart_split(ipt[1] if len(ipt) > 1 else "", subcommand.args_data, subcommand=bool(subcommand.subcommands))
        init_args = (ctx,) if self.addon is None or subcommand_invoking else (self.addon, ctx)
        return await tgt(*init_args, *args, **kwargs)


def command(name: typing.Optional[str] = None, *, aliases: typing.Optional[typing.List[str]] = None):
    def wrap(func):
        return Command(func, name, aliases=aliases)
    return wrap



import typing
from .command import Command
try:
    from dico_interaction import InteractionCommand, ComponentCallback
except ImportError:
    InteractionCommand = None
    ComponentCallback = None

if typing.TYPE_CHECKING:
    from .bot import Bot
    from .context import Context


class Listener:
    def __init__(self, func, event: str):
        self.__func = func
        self.event = event
        self.addon = None

    def register_addon(self, addon):
        self.addon = addon

    async def func(self, *args, **kwargs):
        args = (self.addon, *args)
        return await self.__func(*args, **kwargs)


def on(event: str = None):
    def wrap(func):
        listener = Listener(func, event or func.__name__)
        return listener
    return wrap


on_ = on


class Addon:
    name: str

    def __init_subclass__(cls, **kwargs):
        cls.name = kwargs.get("name", cls.__name__)

    def __init__(self, bot: "Bot"):
        self.bot = bot
        resp = [getattr(self, x) for x in dir(self)]
        self.commands: typing.List[Command] = [x for x in resp if isinstance(x, Command) and not x.is_subcommand]
        self.listeners: typing.List[Listener] = [x for x in resp if isinstance(x, Listener)]
        self.interactions: typing.List["InteractionCommand"] = [x for x in resp if InteractionCommand is not None and isinstance(x, InteractionCommand)]
        self.callbacks: typing.List["ComponentCallback"] = [x for x in resp if ComponentCallback is not None and isinstance(x, ComponentCallback)]
        self.on_load()

    def __str__(self):
        return self.name

    def on_load(self):
        pass

    def on_unload(self):
        pass

    async def addon_check(self, ctx):  # noqa
        return True

    async def on_addon_command_error(self, ctx, ex):  # noqa
        return False

    async def on_addon_interaction_error(self, interaction, ex):  # noqa
        return False
